import base64
import glob
import json
import math
import os

import pytest

import geobuf
from geobuf import Decoder, Encoder

exclude = {'precision.json'}
files = glob.glob(os.path.join(os.path.dirname(__file__), "fixtures/*.json"))
coding_fixtures = [filename for filename in files if os.path.basename(filename) not in exclude]


@pytest.mark.parametrize("filename", coding_fixtures)
def test_coding(filename):
    geojson = json.loads(open(filename).read())
    pb = Encoder().encode(geojson)
    geojson2 = Decoder().decode(pb)
    assert geojson == geojson2


def test_high_precision():
    filename = os.path.join(os.path.dirname(__file__), "fixtures/precision.json")
    with open(filename) as f:
        geojson = json.loads(f.read())
    pbf = Encoder().encode(geojson)
    geojson2 = Decoder().decode(pbf)
    ring = geojson2['features'][0]['geometry']['coordinates'][0]
    assert ring[0] == ring[4]


def test_line_accumulating_error():
    """
    Generate a line of 40 points. Each point's x coordinate, x[n] is at x[n - 1] + 1 + d, where
    d is a floating point number that just rounds to 0 at 6 decimal places, i.e. 0.00000049.
    Therefore a delta compression method that only computes x[n] - x[n - 1] and rounds to 6 d.p.
    will get a constant delta of 1.000000. The result will be an accumulated error along the
    line of 0.00000049 * 40 = 0.0000196 over the full length.
    """
    feature = {
        'type': 'MultiPolygon',
        'coordinates': [[[]]],
    }
    points = 40
    # X coordinates[0, 1.00000049, 2.00000098, 3.00000147, 4.00000196, ...,
    #               37.00001813, 38.00001862, 39.00001911, 40.00001960, 0]
    feature['coordinates'][0][0] = [[i * 1.00000049, 0] for i in range(0, points + 1)]
    feature['coordinates'][0][0].append([0, 0])
    pbf = Encoder().encode(feature)
    round_tripped = Decoder().decode(pbf)

    def round_x(coord):
        x, _ = coord
        return round(x * 1000000) / 1000000.0

    xs_orig = [round_x(coord) for coord in feature['coordinates'][0][0]]
    xs_round_tripped = [round_x(coord) for coord in round_tripped['coordinates'][0][0]]
    assert xs_round_tripped == xs_orig


def test_circle_accumulating_error():
    feature = {
        'type': 'MultiPolygon',
        'coordinates': [[[]]],
    }
    points = 16
    feature['coordinates'][0][0] = [
        [math.cos(math.pi * 2.0 * i / points), math.sin(math.pi * 2.0 * i / points)]
        for i in range(0, points + 1)
    ]
    pbf = Encoder().encode(feature)
    round_tripped = Decoder().decode(pbf)

    def round_coord(coord):
        x, y = coord
        return [round(x * 1000000), round(y * 1000000)]

    ring_orig = [round_coord(coord) for coord in feature['coordinates'][0][0]]
    ring_round_tripped = [round_coord(coord) for coord in round_tripped['coordinates'][0][0]]
    assert ring_round_tripped == ring_orig


def test_dimensions():
    """
    Generate a line of 8 points. Each point is a coordinate with an altitude.
    Ensure dim parameter allows to encode and decode a geometry.

    Using dim=2 will reduce the coordinates by removing the altitude.
    Using dim=3 must conserve the altitude
    """
    feature = {
        'type': 'MultiPolygon',
        'coordinates': [[[]]],
    }
    points = 8
    # Z coordinates[0, 1, 2, 3, ... , 8, 0]
    feature['coordinates'][0][0] = [[0, 0, i] for i in range(0, points + 1)]
    feature['coordinates'][0][0].append([0, 0, 0])

    pbf = Encoder().encode(feature, dim=2)
    dim2 = Decoder().decode(pbf)

    dim2_orig = [[x, y] for x, y, z in feature['coordinates'][0][0]]
    dim2 = dim2['coordinates'][0][0]
    assert dim2 == dim2_orig

    pbf = Encoder().encode(feature, dim=3)
    dim3 = Decoder().decode(pbf)

    dim_orig = feature['coordinates'][0][0]
    dim3 = dim3['coordinates'][0][0]
    assert dim3 == dim_orig


def test_decode_js_encoded_dim_zero():
    """
    Test decoding data where dimensions field is unset (JS encoder behavior).

    The JS geobuf encoder omits the dimensions field when dim=2 (optimization).
    Proto3 returns 0 for unset uint32 fields. The Python decoder must handle
    this by defaulting to 2, matching the JS decoder behavior.

    See: https://github.com/pygeobuf/pygeobuf/issues/59
    """
    # This data was encoded by the JS library (via PostGIS ST_AsGeobuf)
    # with dimensions field omitted (unset, proto3 returns 0)
    data = base64.b64decode('GAAiEAoOCgwIBBoIAAAAAgIAAAE=')
    result = geobuf.decode(data)

    assert result['type'] == 'FeatureCollection'
    assert len(result['features']) == 1
    assert result['features'][0]['geometry']['type'] == 'Polygon'



import os.path

from click.testing import CliRunner
import pytest

import geobuf
from geobuf.scripts.cli import cli


@pytest.fixture
def props_json():
    return open(
        os.path.join(os.path.dirname(__file__), "fixtures/props.json")).read()


def test_cli_version():
    runner = CliRunner()
    result = runner.invoke(cli, ['--version'])
    assert result.output.strip() == geobuf.__version__


def test_cli_encode_err():
    runner = CliRunner()
    result = runner.invoke(cli, ['encode'], "0")
    assert result.exit_code == 1


def test_cli_decode_err():
    runner = CliRunner()
    result = runner.invoke(cli, ['decode'], "0")
    assert result.exit_code == 1


def test_cli_roundtrip(props_json):
    """ tests the roundtrip of encoding and decoding geobuf using the cli, essentially:
        $ geobuf encode < props.json | geobuf decode
    """
    runner = CliRunner()
    result = runner.invoke(cli, ['encode'], input=props_json)
    assert result.exit_code == 0
    pbf = result.stdout_bytes
    result = runner.invoke(cli, ['decode'], input=pbf)
    assert result.exit_code == 0
    assert "@context" in result.output
    assert result.output.count("Feature") == 6



"""
Pygeobuf's command line interface.

`geobuf encode` and `geobuf decode` read and write to the standard IO
streams.

  $ geobuf encode < foo.json > foo.pbf
  $ geobuf decode < foo.pbf > bar.json

"""


import json
import logging
import sys

import click

import geobuf


logging.basicConfig(stream=sys.stderr, level=logging.INFO)


def print_version(ctx, param, value):
    """Print module version and exit."""
    if not value or ctx.resilient_parsing:
        return
    click.echo(geobuf.__version__)
    ctx.exit()


@click.group(help="Pygeobuf command line interface.")
@click.option('--version', is_flag=True, callback=print_version,
              expose_value=False, is_eager=True,
              help="Print Pygeobuf version.")
def cli():
    """Create a context for this command group."""


@cli.command(short_help="Encode a GeoJSON feature collection.")
@click.option('--precision', type=int, default=6,
              help="Decimal encoding precision of coordinates.")
@click.option('--with-z/--without-z', default=False,
              help="Encode Z coordinate values as well as X, Y.")
def encode(precision, with_z):
    """Given GeoJSON on stdin, writes a geobuf file to stdout."""
    logger = logging.getLogger('geobuf')
    stdin = click.get_text_stream('stdin')
    sink = click.get_binary_stream('stdout')
    try:
        data = json.load(stdin)
        pbf = geobuf.encode(
            data,
            precision if precision >= 0 else 6,
            3 if with_z else 2)
        sink.write(pbf)
        sys.exit(0)
    except Exception:
        logger.exception("Failed. Exception caught")
        sys.exit(1)


@cli.command(short_help="Decode a Geobuf byte string.")
def decode():
    """Given a Geobuf byte string on stdin, write a GeoJSON feature
    collection to stdout."""
    logger = logging.getLogger('geobuf')
    stdin = click.get_binary_stream('stdin')
    sink = click.get_text_stream('stdout')
    try:
        pbf = stdin.read()
        data = geobuf.decode(pbf)
        json.dump(data, sink)
        sys.exit(0)
    except Exception:
        logger.exception("Failed. Exception caught")
        sys.exit(1)



# -*- coding: utf-8 -*-

import collections
import json

from . import geobuf_pb2


class Decoder:
    geometry_types = ('Point', 'MultiPoint', 'LineString', 'MultiLineString',
                      'Polygon', 'MultiPolygon', 'GeometryCollection')

    def __init__(self):
        self.data: geobuf_pb2.Data = geobuf_pb2.Data()
        self.e: int = 1
        self.dim: int = 2

    def decode(self, data_str: bytes):

        data = self.data = geobuf_pb2.Data()
        data.ParseFromString(data_str)

        self.e = pow(10, data.precision)
        self.dim = data.dimensions or 2

        data_type = data.WhichOneof('data_type')

        if data_type == 'feature_collection':
            return self.decode_feature_collection(data.feature_collection)
        elif data_type == 'feature':
            return self.decode_feature(data.feature)
        elif data_type == 'geometry':
            return self.decode_geometry(data.geometry)

    def decode_feature_collection(self, feature_collection):
        obj = {'type': 'FeatureCollection', 'features': []}
        self.decode_properties(feature_collection.custom_properties, feature_collection.values, obj)
        for feature in feature_collection.features:
            obj['features'].append(self.decode_feature(feature))

        return obj

    def decode_feature(self, feature):
        obj = collections.OrderedDict()
        obj['type'] = 'Feature'

        self.decode_properties(feature.custom_properties, feature.values, obj)

        self.decode_id(feature, obj)
        obj['geometry'] = self.decode_geometry(feature.geometry)
        if len(feature.properties):
            obj['properties'] = self.decode_properties(feature.properties, feature.values)

        return obj

    def decode_properties(self, props, values, dest=None):
        if dest is None:
            dest = {}
        for i in range(0, len(props), 2):
            key = self.data.keys[props[i]]
            val = values[props[i + 1]]

            value_type = val.WhichOneof('value_type')
            if value_type == 'string_value':
                dest[key] = val.string_value
            elif value_type == 'double_value':
                dest[key] = val.double_value
            elif value_type == 'pos_int_value':
                dest[key] = val.pos_int_value
            elif value_type == 'neg_int_value':
                dest[key] = -val.neg_int_value
            elif value_type == 'bool_value':
                dest[key] = val.bool_value
            elif value_type == 'json_value':
                dest[key] = json.loads(val.json_value)
        return dest

    @staticmethod
    def decode_id(obj, obj_json):
        id_type = obj.WhichOneof('id_type')
        if id_type == 'id':
            obj_json['id'] = obj.id
        elif id_type == 'int_id':
            obj_json['id'] = obj.int_id

    def decode_geometry(self, geometry):
        obj = collections.OrderedDict()
        gt = obj['type'] = self.geometry_types[geometry.type]

        self.decode_properties(geometry.custom_properties, geometry.values, obj)

        if gt == 'GeometryCollection':
            obj['geometries'] = [self.decode_geometry(geom) for geom in geometry.geometries]
        elif gt == 'Point':
            obj['coordinates'] = self.decode_point(geometry.coords)
        elif gt == 'MultiPoint':
            obj['coordinates'] = self.decode_line(geometry.coords)
        elif gt == 'LineString':
            obj['coordinates'] = self.decode_line(geometry.coords)
        elif gt == 'MultiLineString':
            obj['coordinates'] = self.decode_multi_line(geometry)
        elif gt == 'Polygon':
            obj['coordinates'] = self.decode_multi_line(geometry, is_closed=True)
        elif gt == 'MultiPolygon':
            obj['coordinates'] = self.decode_multi_polygon(geometry)

        return obj

    def decode_coord(self, coord):
        return float(coord) / self.e

    def decode_point(self, coords):
        return [self.decode_coord(x) for x in coords]

    def decode_line(self, coords, is_closed=False):
        obj = []

        r = range(self.dim)
        r2 = range(0, len(coords), self.dim)
        p0 = [0] * self.dim
        for i in r2:
            p = [p0[j] + coords[i + j] for j in r]
            obj.append(self.decode_point(p))
            p0 = p

        if is_closed:
            p = [coords[j] for j in r]
            obj.append(self.decode_point(p))

        return obj

    def decode_multi_line(self, geometry, is_closed=False):
        if len(geometry.lengths) == 0:
            return [self.decode_line(geometry.coords, is_closed=is_closed)]

        obj = []
        i = 0

        for l in geometry.lengths:
            obj.append(self.decode_line(geometry.coords[i:i + l * self.dim], is_closed=is_closed))
            i += l * self.dim

        return obj

    def decode_multi_polygon(self, geometry):
        if len(geometry.lengths) == 0:
            return [[self.decode_line(geometry.coords, is_closed=True)]]

        obj = []
        i = 0
        j = 1
        num_polygons = geometry.lengths[0]

        for n in range(num_polygons):  # for every polygon
            num_rings = geometry.lengths[j]
            j += 1
            rings = []
            for l in geometry.lengths[j:j + num_rings]:
                rings.append(self.decode_line(geometry.coords[i:i + l * self.dim], is_closed=True))
                j += 1
                i += l * self.dim
            obj.append(rings)
        return obj



# -*- coding: utf-8 -*-

import collections
import json

from typing import Mapping
import six

from . import geobuf_pb2


class Encoder:
    geometry_types = {
        'Point': 0,
        'MultiPoint': 1,
        'LineString': 2,
        'MultiLineString': 3,
        'Polygon': 4,
        'MultiPolygon': 5,
        'GeometryCollection': 6,
    }

    def __init__(self):
        self.json: Mapping = dict()
        self.data: geobuf_pb2.Data = geobuf_pb2.Data()
        self.precision: int = 6
        self.dim: int = 2
        self.e: int = pow(10, self.precision)
        self.keys = collections.OrderedDict()

    def encode(self, data_json: Mapping, precision: int = 6, dim: int = 2):
        obj = self.json = data_json
        data = self.data = geobuf_pb2.Data()
        data.dimensions = dim
        data.precision = precision

        self.precision = precision
        self.dim = dim
        self.e = pow(10, precision)  # multiplier for converting coordinates into integers

        data_type = obj['type']

        if data_type == 'FeatureCollection':
            self.encode_feature_collection(data.feature_collection, obj)
        elif data_type == 'Feature':
            self.encode_feature(data.feature, obj)
        else:
            self.encode_geometry(data.geometry, obj)

        return data.SerializeToString()

    def encode_feature_collection(self, feature_collection, feature_collection_json):
        self.encode_custom_properties(feature_collection, feature_collection_json, ('type', 'features'))
        for feature_json in feature_collection_json.get('features'):
            self.encode_feature(feature_collection.features.add(), feature_json)

    def encode_feature(self, feature, feature_json):
        self.encode_id(feature, feature_json.get('id'))
        self.encode_properties(feature, feature_json.get('properties'))
        self.encode_custom_properties(feature, feature_json, ('type', 'id', 'properties', 'geometry'))
        self.encode_geometry(feature.geometry, feature_json.get('geometry'))

    def encode_geometry(self, geometry, geometry_json):

        gt = geometry_json['type']
        coords = geometry_json.get('coordinates')

        geometry.type = self.geometry_types[gt]

        self.encode_custom_properties(geometry, geometry_json,
                                      ('type', 'id', 'coordinates', 'arcs', 'geometries', 'properties'))

        if gt == 'GeometryCollection':
            for geom in geometry_json.get('geometries'):
                self.encode_geometry(geometry.geometries.add(), geom)
        elif gt == 'Point':
            self.add_point(geometry.coords, coords)
        elif gt == 'MultiPoint':
            self.add_line(geometry.coords, coords)
        elif gt == 'LineString':
            self.add_line(geometry.coords, coords)
        elif gt == 'MultiLineString':
            self.add_multi_line(geometry, coords)
        elif gt == 'Polygon':
            self.add_multi_line(geometry, coords, is_closed=True)
        elif gt == 'MultiPolygon':
            self.add_multi_polygon(geometry, coords)

    def encode_properties(self, obj, props_json):
        if props_json:
            for key, val in props_json.items():
                self.encode_property(key, val, obj.properties, obj.values)

    def encode_custom_properties(self, obj, obj_json, exclude):
        for key, val in obj_json.items():
            if key not in exclude:
                self.encode_property(key, val, obj.custom_properties, obj.values)

    def encode_property(self, key, val, properties, values):
        keys = self.keys

        if key not in keys:
            keys[key] = True
            self.data.keys.append(key)
            key_index = len(self.data.keys) - 1
        else:
            key_index = list(keys.keys()).index(key)

        value = values.add()

        if isinstance(val, dict) or isinstance(val, list):
            value.json_value = json.dumps(val, separators=(',', ':')).encode('utf-8')
        elif isinstance(val, six.text_type):
            value.string_value = val
        elif isinstance(val, float):
            if val.is_integer():
                self.encode_int(int(val), value)
            else:
                value.double_value = val
        elif isinstance(val, bool):
            value.bool_value = val
        elif isinstance(val, six.integer_types):
            self.encode_int(val, value)

        properties.append(key_index)
        properties.append(len(values) - 1)

    @staticmethod
    def encode_int(val, value):
        try:
            if val >= 0:
                value.pos_int_value = val
            else:
                value.neg_int_value = -val
        except ValueError:
            value.double_value = val

    @staticmethod
    def encode_id(obj, id):
        if id is not None:
            if isinstance(id, int):
                obj.int_id = id
            else:
                obj.id = str(id)

    def add_coord(self, coords, coord):
        coords.append(int(round(coord * self.e)))

    def add_point(self, coords, point):
        for x in point:
            self.add_coord(coords, x)

    def add_line(self, coords, points, is_closed=False):
        sum = [0] * self.dim
        r = range(0, len(points) - int(is_closed))
        for i in r:
            for j in range(0, self.dim):
                n = int(round(points[i][j] * self.e) - sum[j])
                coords.append(n)
                sum[j] += n

    def add_multi_line(self, geometry, lines, is_closed=False):
        if len(lines) != 1:
            for points in lines:
                geometry.lengths.append(len(points) - int(is_closed))

        for points in lines:
            self.add_line(geometry.coords, points, is_closed)

    def add_multi_polygon(self, geometry, polygons):
        if len(polygons) != 1 or len(polygons[0]) != 1:
            geometry.lengths.append(len(polygons))
            for rings in polygons:
                geometry.lengths.append(len(rings))
                for points in rings:
                    geometry.lengths.append(len(points) - 1)

        for rings in polygons:
            for points in rings:
                self.add_line(geometry.coords, points, is_closed=True)



from setuptools import setup, find_packages
from pathlib import Path

setup(name='geobuf',
      version='2.0.1',
      description=(
          u"Geobuf is a compact binary geospatial format for lossless "
          u"compression of GeoJSON."),
      long_description=Path('README.md').read_text(),
      long_description_content_type='text/markdown',
      classifiers=[],
      keywords='data gis geojson protobuf',
      author=u"Vladimir Agafonkin",
      author_email='vladimir@mapbox.com',
      url='https://github.com/mapbox/pygeobuf',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=['click', 'protobuf', 'six'],
      extras_require={
          'test': ['pytest'],
      },
      entry_points="""
      [console_scripts]
      geobuf=geobuf.scripts.cli:cli
      """)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: geobuf.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cgeobuf.proto\"\xde\x07\n\x04\x44\x61ta\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\x12\n\ndimensions\x18\x02 \x01(\r\x12\x11\n\tprecision\x18\x03 \x01(\r\x12\x35\n\x12\x66\x65\x61ture_collection\x18\x04 \x01(\x0b\x32\x17.Data.FeatureCollectionH\x00\x12 \n\x07\x66\x65\x61ture\x18\x05 \x01(\x0b\x32\r.Data.FeatureH\x00\x12\"\n\x08geometry\x18\x06 \x01(\x0b\x32\x0e.Data.GeometryH\x00\x1a\xa2\x01\n\x07\x46\x65\x61ture\x12 \n\x08geometry\x18\x01 \x01(\x0b\x32\x0e.Data.Geometry\x12\x0c\n\x02id\x18\x0b \x01(\tH\x00\x12\x10\n\x06int_id\x18\x0c \x01(\x12H\x00\x12\x1b\n\x06values\x18\r \x03(\x0b\x32\x0b.Data.Value\x12\x12\n\nproperties\x18\x0e \x03(\r\x12\x19\n\x11\x63ustom_properties\x18\x0f \x03(\rB\t\n\x07id_type\x1a\xdd\x02\n\x08Geometry\x12!\n\x04type\x18\x01 \x01(\x0e\x32\x13.Data.Geometry.Type\x12\x0f\n\x07lengths\x18\x02 \x03(\r\x12\x0e\n\x06\x63oords\x18\x03 \x03(\x12\x12\"\n\ngeometries\x18\x04 \x03(\x0b\x32\x0e.Data.Geometry\x12\x1b\n\x06values\x18\r \x03(\x0b\x32\x0b.Data.Value\x12\x19\n\x11\x63ustom_properties\x18\x0f \x03(\r\"\xb0\x01\n\x04Type\x12\x0e\n\nTYPE_EMPTY\x10\x00\x12\x0e\n\nTYPE_POINT\x10\x01\x12\x13\n\x0fTYPE_MULTIPOINT\x10\x02\x12\x13\n\x0fTYPE_LINESTRING\x10\x03\x12\x18\n\x14TYPE_MULTILINESTRING\x10\x04\x12\x10\n\x0cTYPE_POLYGON\x10\x05\x12\x15\n\x11TYPE_MULTIPOLYGON\x10\x06\x12\x1b\n\x17TYPE_GEOMETRYCOLLECTION\x10\x07\x1al\n\x11\x46\x65\x61tureCollection\x12\x1f\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32\r.Data.Feature\x12\x1b\n\x06values\x18\r \x03(\x0b\x32\x0b.Data.Value\x12\x19\n\x11\x63ustom_properties\x18\x0f \x03(\r\x1a\xa3\x01\n\x05Value\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x16\n\x0c\x64ouble_value\x18\x02 \x01(\x01H\x00\x12\x17\n\rpos_int_value\x18\x03 \x01(\x04H\x00\x12\x17\n\rneg_int_value\x18\x04 \x01(\x04H\x00\x12\x14\n\nbool_value\x18\x05 \x01(\x08H\x00\x12\x14\n\njson_value\x18\x06 \x01(\x0cH\x00\x42\x0c\n\nvalue_typeB\x0b\n\tdata_typeb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'geobuf_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_DATA']._serialized_start=17
  _globals['_DATA']._serialized_end=1007
  _globals['_DATA_FEATURE']._serialized_start=204
  _globals['_DATA_FEATURE']._serialized_end=366
  _globals['_DATA_GEOMETRY']._serialized_start=369
  _globals['_DATA_GEOMETRY']._serialized_end=718
  _globals['_DATA_GEOMETRY_TYPE']._serialized_start=542
  _globals['_DATA_GEOMETRY_TYPE']._serialized_end=718
  _globals['_DATA_FEATURECOLLECTION']._serialized_start=720
  _globals['_DATA_FEATURECOLLECTION']._serialized_end=828
  _globals['_DATA_VALUE']._serialized_start=831
  _globals['_DATA_VALUE']._serialized_end=994
# @@protoc_insertion_point(module_scope)
